@@ -1,6 +1,7 @@
-{-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE FlexibleInstances, UndecidableInstances #-}
+{-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE MultiParamTypeClasses      #-}
+{-# LANGUAGE UndecidableInstances       #-}
 {-# OPTIONS -fno-warn-orphans #-}
 
 {- |
@@ -35,26 +36,34 @@ module Control.Monad.Random (
     runRand,
     evalRandIO,
     fromList,
-    Rand, RandT -- but not the data constructors
+    uniform,
+    Rand, RandT, -- but not the data constructors
+    -- * Special lift functions
+    liftRand,
+    liftRandT
     -- * Example
     -- $RandExample
     ) where
 
-import System.Random
-import Control.Monad()
-import Control.Monad.Identity
-import Control.Monad.Random.Class
-import Control.Monad.Reader
-import Control.Monad.State
-import Control.Monad.Trans()
-import Control.Monad.Writer
-import Control.Arrow
-import Control.Applicative
+import           Control.Applicative
+import           Control.Arrow
+import           Control.Monad              ()
+import           Control.Monad.Cont
+import           Control.Monad.Error
+import           Control.Monad.Identity
+import           Control.Monad.Random.Class
+import           Control.Monad.Reader
+import           Control.Monad.State
+import           Control.Monad.Trans        ()
+import           Control.Monad.Trans.Identity
+import           Control.Monad.Trans.Maybe
+import           Control.Monad.Writer
+import           System.Random
 
 -- | A monad transformer which adds a random number generator to an
 -- existing monad.
-newtype (RandomGen g) => RandT g m a = RandT (StateT g m a)
-    deriving (Functor, Monad, MonadTrans, MonadIO, MonadFix)
+newtype RandT g m a = RandT (StateT g m a)
+    deriving (Functor, Monad, MonadTrans, MonadIO, MonadFix, MonadReader r, MonadWriter w)
 
 instance (Functor m,Monad m) => Applicative (RandT g m) where
   pure = return
@@ -66,6 +75,18 @@ liftState t = do v <- get
                  put v'
                  return x
 
+-- | Lift arbitrary action to RandT
+liftRandT :: (Monad m, RandomGen g, Random a) =>
+             (g -> m (a, g)) -- ^ action returning value and new generator state
+             -> RandT g m a
+liftRandT = RandT . StateT
+
+-- | Lift arbitrary action to Rand
+liftRand :: (RandomGen g, Random a) =>
+            (g -> (a, g)) -- ^ action returning value and new generator state
+            -> Rand g a
+liftRand = Rand . RandT . liftState
+
 instance (Monad m, RandomGen g) => MonadRandom (RandT g m) where
     getRandom = RandT . liftState $ random
     getRandoms = RandT . liftState $ first randoms . split
@@ -102,10 +123,9 @@ evalRand (Rand x) g = runIdentity (evalRandT x g)
 runRand :: (RandomGen g) => Rand g a -> g -> (a, g)
 runRand (Rand x) g = runIdentity (runRandT x g)
 
--- | Evaluate a random computation in the IO monad, using the random number
--- generator supplied by 'System.Random.getStdRandom'.
+-- | Evaluate a random computation in the IO monad, splitting the global standard generator to get a new one for the computation.
 evalRandIO :: Rand StdGen a -> IO a
-evalRandIO (Rand (RandT x)) = getStdRandom (runIdentity . runStateT x)
+evalRandIO x = fmap (evalRand x) newStdGen
 
 -- | Sample a random value from a weighted list.  The total weight of all
 -- elements must not be 0.
@@ -120,6 +140,16 @@ fromList xs = do
   p <- liftM toRational $ getRandomR (0.0,s)
   return . fst . head $ dropWhile (\(_,q) -> q < p) cs
 
+-- | Sample a value from a uniform distribution of a list of elements.
+uniform :: (MonadRandom m) => [a] -> m a
+uniform = fromList . fmap (flip (,) 1)
+
+instance (MonadRandom m) => MonadRandom (IdentityT m) where
+    getRandom = lift getRandom
+    getRandomR = lift . getRandomR
+    getRandoms = lift getRandoms
+    getRandomRs = lift . getRandomRs
+
 instance (MonadRandom m) => MonadRandom (StateT s m) where
     getRandom = lift getRandom
     getRandomR = lift . getRandomR
@@ -138,6 +168,27 @@ instance (MonadRandom m) => MonadRandom (ReaderT r m) where
     getRandoms = lift getRandoms
     getRandomRs = lift . getRandomRs
 
+instance (Error e, MonadRandom m) => MonadRandom (ErrorT e m) where
+    getRandom = lift getRandom
+    getRandomR = lift . getRandomR
+    getRandoms = lift getRandoms
+    getRandomRs = lift . getRandomRs
+
+instance (MonadRandom m) => MonadRandom (MaybeT m) where
+    getRandom = lift getRandom
+    getRandomR = lift . getRandomR
+    getRandoms = lift getRandoms
+    getRandomRs = lift . getRandomRs
+
+instance MonadRandom m => MonadRandom (ContT r m) where
+    getRandom = lift getRandom
+    getRandomR = lift . getRandomR
+    getRandoms = lift getRandoms
+    getRandomRs = lift . getRandomRs
+
+instance (MonadSplit g m) => MonadSplit g (IdentityT m) where
+    getSplit = lift getSplit
+
 instance (MonadSplit g m) => MonadSplit g (StateT s m) where
     getSplit = lift getSplit
 
@@ -147,19 +198,19 @@ instance (MonadSplit g m, Monoid w) => MonadSplit g (WriterT w m) where
 instance (MonadSplit g m) => MonadSplit g (ReaderT r m) where
     getSplit = lift getSplit
 
+instance (Error e, MonadSplit g m) => MonadSplit g (ErrorT e m) where
+    getSplit = lift getSplit
+
+instance (MonadSplit g m) => MonadSplit g (MaybeT m) where
+    getSplit = lift getSplit
+
+instance (MonadSplit g m) => MonadSplit g (ContT r m) where
+    getSplit = lift getSplit
+
 instance (MonadState s m, RandomGen g) => MonadState s (RandT g m) where
     get = lift get
     put = lift . put
 
-instance (MonadReader r m, RandomGen g) => MonadReader r (RandT g m) where
-    ask = lift ask
-    local f (RandT m) = RandT $ local f m
-
-instance (MonadWriter w m, RandomGen g, Monoid w) => MonadWriter w (RandT g m) where
-    tell = lift . tell
-    listen (RandT m) = RandT $ listen m
-    pass (RandT m) = RandT $ pass m
-
 instance MonadRandom IO where
     getRandom = randomIO
     getRandomR = randomRIO
